﻿using System.Windows.Forms;
using BlockMania.GeneralStates.Screen_01_Home;
using BlockMania.GeneralStates.Screen_02_Play;
using BlockMania.GeneralStates.Screen_03_Settings;
using BlockMania.GeneralStates.Screen_04_Help;
using BlockMania.Util;
using BlockMania.Util.BlockTools;
using BlockMania.View;

namespace BlockMania.Global
{
	public enum Screen
	{
		HOME,
		PLAY,
		SETTINGS,
		HELP
	}

	public enum UserInput
	{
		NONE,
		ESCAPE,

		SWITCH_TO_HOME,
		SWITCH_TO_PLAY,
		SWITCH_TO_SETTINGS,
		SWITCH_TO_HELP,

		MOVE_LEFT,
		MOVE_RIGHT,
		MOVE_UP,
		MOVE_DOWN,

		ROTATE,
		FLIP,

		SELECT_BLOCK,
		 F5,
		 F6,
		 F7,
		 F8,
		 F9
	}

	internal static class GlobalController
	{
		private static readonly RepetitionChecker repetitionChecker = new RepetitionChecker();
		private static UserInput userInput;

		private static Screen activeScreen;

		internal static void Init(Panel p)
		{
			GlobalCanvas.Init(p);

			ResourceLocator.Init();
			SettingStorage.ReadSettingsFromDisk();

			ColorManager.Init();
			FontManager.Init();

			PreviewField.SetRandomBlock();
			Gamefield.Init();

			Screen_1_Home.Init();
			Screen_2_Play.Init();
			Screen_3_Settings.Init();
			Screen_4_Help.Init();

			BlockSpeed.Init();
			ScoreManager.Init();

			activeScreen = Screen.HOME;
			GameLoopTimer.Start();
		}

		internal static void SetUserInput(UserInput input)
		{
			userInput = input;
		}

		internal static void HandleGameTick()
		{
			UserInput input = userInput; // Defensive copy, this copy will keep its value even after userInput changes.

			bool isNotRepeating = repetitionChecker.IsNotRepeating(input);
			HandleInput(input, isNotRepeating);

			GlobalCanvas.DrawBackgroundColor();
			DistributeGameTick(input, isNotRepeating);
			GlobalCanvas.Render();
		}

		private static void HandleInput(UserInput input, bool isNotRepeating)
		{
			if (input != UserInput.NONE && isNotRepeating)
			{
				switch (input)
				{
					case UserInput.SWITCH_TO_HOME     : SwitchToHome()     ; break;
					case UserInput.SWITCH_TO_SETTINGS : SwitchToSettings() ; break;
					case UserInput.SWITCH_TO_HELP     : SwitchToHelp()     ; break;
					case UserInput.ESCAPE             : TogglePause()      ; break;
				}
			}
		}

		internal static void SwitchToHome()
		{
			Screen_1_Home.Reset();
			activeScreen = Screen.HOME;
		}

		internal static void SwitchToPlay()
		{
			Screen_2_Play.Reset();
			activeScreen = Screen.PLAY;
		}

		private static void SwitchToSettings()
		{
			activeScreen = Screen.SETTINGS;
		}

		private static void SwitchToHelp()
		{
			activeScreen = Screen.HELP;
		}

		private static void TogglePause()
		{
			if (activeScreen == Screen.PLAY)
			{
				Screen_2_Play.TogglePause();
			}
		}

		private static void DistributeGameTick(UserInput input, bool isNotRepeating)
		{
			switch (activeScreen)
			{
				case Screen.HOME     : Screen_1_Home.HandleGameTick(     input, isNotRepeating) ; break;
				case Screen.PLAY     : Screen_2_Play.HandleGameTick(     input, isNotRepeating) ; break;
				case Screen.SETTINGS : Screen_3_Settings.HandleGameTick( input, isNotRepeating) ; break;
				case Screen.HELP     : Screen_4_Help.HandleGameTick(     input, isNotRepeating) ; break;
			}
		}

	}
}