﻿using System;
using System.Runtime.InteropServices;
using BlockMania.Util.PrecisionTiming.Base;
using BlockMania.Util.PrecisionTiming.Exception;
using BlockMania.Util.PrecisionTiming.Multimedia;
using static BlockMania.Util.PrecisionTiming.Multimedia.MMTimerExports;

namespace BlockMania.Util.PrecisionTiming
{
	/// <summary>
	/// High Resolution Multimedia Timer Wrapper
	/// <para>The default timer will Fire the Tick event every 1 Millisecond forever</para>
	/// <para>Subscribe to <see cref="EventHandler"/> <see cref="Tick"/></para>
	/// <para>You can also provide an Action using <see cref="SetAction"/> or <see cref="SetInterval(Action, int, bool, int, TimerMode, bool)"/></para>
	/// </summary>
	public class PrecisionTimer : IDisposable
	{
		/// <summary>
		/// Occurs when the <see cref="PrecisionTimer"/> has started.
		/// </summary>
		public volatile EventHandler Started;

		/// <summary>
		/// Occurs when the <see cref="PrecisionTimer"/> has stopped.
		/// </summary>
		public volatile EventHandler Stopped;

		/// <summary>
		/// Occurs when the <see cref="PrecisionTimer"/> Ticks at the Interval.
		/// </summary>
		public volatile EventHandler Tick;

		#region [ Members ]

		private volatile int m_timerID; // Timer identifier.
		private volatile int m_period; // Period between timer events in milliseconds.
		private volatile int m_resolution; // Timer resolution in milliseconds.
		private volatile bool m_running; // Indicates whether or not the timer is running.
		private volatile bool m_disposed; // Indicates whether or not the timer has been disposed.
		private volatile object m_lock = new object(); // Lock
		private volatile EventArgs m_eventArgs; // Private user event args to pass into Ticks call
		private volatile Action m_action = null; // The Action
		private volatile TimerProc m_timeProc; // Called by Windows when a timer periodic event occurs.
		private volatile TimerMode m_mode; // Timer mode.
		private volatile bool m_noTick = false; // No Tick Event
		private volatile SemaphoreWeak m_slimWeak = new SemaphoreWeak(); // Weak Slim (Minimum Delay makes this possible)

		#endregion [ Members ]

		#region [ Constructors / Destructor ]

		/// <summary>
		/// Create a PrecisionTimer and tell the Garbage Collector to leave it alone
		/// </summary>
		public PrecisionTimer()
		{
			GC.SuppressFinalize(this);
			timeGetDevCaps(ref Capabilities, Marshal.SizeOf(Capabilities));

			// Initialize timer with default values.
			m_mode = TimerMode.Periodic;
			m_running = false;
			m_period = Capabilities.PeriodMinimum;
			m_resolution = Capabilities.PeriodMinimum;
		}

		/// <summary>
		/// Deconstruct
		/// </summary>
		~PrecisionTimer()
		{
			Dispose(false);
			GC.Collect();
		}

		#endregion [ Constructors / Destructor ]

		#region [ Callbacks ]

		internal void TimerCallback(int hwnd, int uMsg, IntPtr passThePointer, int dwTime, int WTF)
		{
			if (m_action != null)
			{
				m_action();
			}

			if (Tick is object)
			{
				Tick(this, m_eventArgs);
			}
		}

		internal void TimerCallbackNoTick(int hwnd, int uMsg, IntPtr passThePointer, int dwTime, int WTF)
		{
			if (m_action != null)
			{
				m_action();
			}
		}

		internal void TimerCallbackSemaphore(int hwnd, int uMsg, IntPtr passThePointer, int dwTime, int WTF)
		{
			if (m_action != null)
			{
				if (m_slimWeak.CanEnter())
				{
					m_action();
					m_slimWeak.Release();
				}
			}

			if (Tick is object)
			{
				Tick(this, m_eventArgs);
			}
		}

		internal void TimerCallbackSemaphoreNoTick(int hwnd, int uMsg, IntPtr passThePointer, int dwTime, int WTF)
		{
			if (m_action != null)
			{
				if (m_slimWeak.CanEnter())
				{
					m_action();
					m_slimWeak.Release();
				}
			}
		}

		internal void TimerCallbackOneShot(int hwnd, int uMsg, IntPtr passThePointer, int dwTime, int WTF)
		{
			lock (m_lock)
			{
				if (m_action != null)
				{
					m_action();
				}

				if (Tick is object)
				{
					Tick(this, m_eventArgs);
				}

				Stop();
			}
		}

		internal void TimerCallbackOneShotNoTick(int hwnd, int uMsg, IntPtr passThePointer, int dwTime, int WTF)
		{
			lock (m_lock)
			{
				if (m_action != null)
				{
					m_action();
				}

				Stop();
			}
		}

		#endregion [ Callbacks ]

		/// <summary>
		/// True if the Timer is running
		/// Will also be false if no <see cref="PrecisionTimer"/> is configured
		/// </summary>
		/// <returns></returns>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		public bool IsRunning()
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedIsRunningString);
			}

			return m_running;
		}

		/// <summary>
		/// Get whether a Tick Event should be used
		/// </summary>
		/// <returns></returns>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		public bool GetNoTick()
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedPeriodString);
			}

			lock (m_lock)
			{
				return m_noTick;
			}
		}

		/// <summary>
		/// Set whether a Tick Event should be used
		/// </summary>
		/// <param name="value"></param>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		/// <exception cref="TimerRunningException">Attempted to configure a timer that was running</exception>
		public void SetNoTick(bool value)
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedPeriodString);
			}

			if (m_running)
			{
				throw new TimerRunningException(Constants.timerRunningString);
			}

			lock (m_lock)
			{
				m_noTick = value;
			}
		}

		/// <summary>
		/// Get the Current Interval (Period)
		/// </summary>
		/// <returns></returns>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		public int GetPeriod()
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedPeriodString);
			}

			lock (m_lock)
			{
				return m_period;
			}
		}

		/// <summary>
		/// Set the Current Interval (Period)
		/// </summary>
		/// <param name="value"></param>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		/// <exception cref="TimerRunningException">Attempted to configure a timer that was running</exception>
		/// <exception cref="ArgumentOutOfRangeException"></exception>
		public void SetPeriod(int value)
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedPeriodString);
			}

			if (m_running)
			{
				throw new TimerRunningException(Constants.timerRunningString);
			}

			if (value > Capabilities.PeriodMaximum)
			{
				throw new ArgumentOutOfRangeException(nameof(value), value, Constants.outOfRangeMaxPeriodString + Capabilities.PeriodMaximum);
			}

			if (value < Capabilities.PeriodMinimum)
			{
				throw new ArgumentOutOfRangeException(nameof(value), value, Constants.outOfRangeMinPeriodString + Capabilities.PeriodMinimum);
			}

			lock (m_lock)
			{
				m_period = value;
			}
		}

		/// <summary>
		/// Get the Current Resolution
		/// </summary>
		/// <returns></returns>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		public int GetResolution()
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedResolutionString);
			}

			lock (m_lock)
			{
				return m_resolution;
			}
		}

		/// <summary>
		/// Set the Current Resolution
		/// </summary>
		/// <param name="value">Desired resolution in milliseconds</param>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		/// <exception cref="TimerRunningException">Attempted to configure a timer that was running</exception>
		/// <exception cref="ArgumentOutOfRangeException">Desired resolution is out of the range of the devcaps</exception>
		public void SetResolution(int value)
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedResolutionString);
			}

			if (m_running)
			{
				throw new TimerRunningException(Constants.timerRunningString);
			}

			if (value > Capabilities.PeriodMaximum)
			{
				throw new ArgumentOutOfRangeException(nameof(value), value, Constants.outOfRangeMaxResolutionString + Capabilities.PeriodMaximum);
			}

			if (value < 0)
			{
				throw new ArgumentOutOfRangeException(nameof(value), value, Constants.outOfRangeMinResolutionString);
			}

			lock (m_lock)
			{
				m_resolution = value;
			}
		}

		/// <summary>
		/// Get the Current User supplied EventArgs
		/// </summary>
		/// <returns></returns>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		public EventArgs GetEventArgs()
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedEventArgsString);
			}

			lock (m_lock)
			{
				return m_eventArgs;
			}
		}

		/// <summary>
		/// Set the User supplied EventArgs
		/// </summary>
		/// <param name="value"></param>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		/// <exception cref="PeriodicTimerRunningException">Attempted to configure a pediodic timer that was running</exception>
		public void SetEventArgs(EventArgs value)
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedEventArgsString);
			}

			if (m_running)
			{
				throw new PeriodicTimerRunningException(Constants.timerRunningString);
			}

			if (value != null)
			{
				lock (m_lock)
				{
					m_eventArgs = value;
				}
			}
		}

		/// <summary>
		/// Get the Current Timer Auto Reset Mode (Periodic/Oneshot)
		/// <para>True for periodic</para>
		/// <para>False for one-shot</para>
		/// </summary>
		/// <returns></returns>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		public bool GetAutoResetMode()
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedAutoResetString);
			}

			lock (m_lock)
			{
				return m_mode == TimerMode.Periodic || m_mode == TimerMode.PeriodicSemaphore;
			}
		}

		/// <summary>
		/// Set the Current Timer Auto Reset Mode (Periodic/Oneshot)
		/// <para>True for periodic</para>
		/// <para>False for one-shot</para>
		/// </summary>
		/// <param name="value"></param>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		/// <exception cref="TimerRunningException">Attempted to configure a timer that was running</exception>
		public void SetAutoResetMode(bool value)
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedAutoResetString);
			}

			if (m_running)
			{
				throw new TimerRunningException(Constants.timerRunningString);
			}

			lock (m_lock)
			{
				m_mode = value ? TimerMode.Periodic : TimerMode.OneShot;
			}
		}

		/// <summary>
		/// Set the Action of the <see cref="PrecisionTimer"/> before you Start the Timer
		/// <para>You could also subscribe to the Event <see cref="Tick"/></para>
		/// </summary>
		/// <param name="TimerTask">The Action</param>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		/// <exception cref="PeriodicTimerRunningException">Attempted to configure a pediodic timer that was running</exception>
		public void SetAction(Action TimerTask)
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedAutoResetString);
			}

			if (m_running)
			{
				throw new PeriodicTimerRunningException(Constants.timerRunningString);
			}

			lock (m_lock)
			{
				m_action = TimerTask;
			}
		}

		/// <summary>
		/// Set the Interval and Tick Event of the <see cref="PrecisionTimer"/> and decide if it should start automatically
		/// <para>Configures the timer with the default settings unless you change them</para>
		/// <para>If you call this with <see href="start = true"/> on a timer that is already running then it will restart, you can check <see cref="IsRunning()"/> to be sure if you don't want this to happen</para>
		/// </summary>
		/// <param name="timerTask">The Action</param>
		/// <param name="interval">The Interval for the TimerTask in Milliseconds</param>
		/// <param name="mode">Timer Mode</param>
		/// <param name="resolution">Resolution for Events in milliseconds - 0 is best available on platform - 1 uses even less CPU</param>
		/// <param name="start">True if the timer should start automatically with the default settings, false if you are going to configure/start it later</param>
		/// <param name="noTick">true if the Tick event should be removed from the Timer Callback</param>
		/// <exception cref="TimerStartException">Timer failed to start</exception>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		public void SetInterval(Action timerTask = null, int interval = 1, bool start = true, int resolution = 0, TimerMode mode = TimerMode.Periodic, bool noTick = false)
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedStartString);
			}

			lock (m_lock)
			{
				m_action = timerTask;
				m_mode = mode;
				m_resolution = resolution;
				m_period = interval;
				m_noTick = noTick;

				if (start)
				{
					if (!m_running)
					{
						Start();
					}
					else
					{
						// Restart
						Stop();
						Start();
					}
				}
			}
		}

		/// <summary>
		/// Start the <see cref="PrecisionTimer"/>
		/// </summary>
		/// <returns></returns>
		/// <exception cref="ObjectDisposedException">Timer was already disposed</exception>
		/// <exception cref="TimerStartException">Timer failed to start</exception>
		public bool Start()
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedStartString);
			}

			lock (m_lock)
			{
				if (m_running)
				{
					return true;
				}

				if (m_mode == TimerMode.Periodic)
				{
					if (!m_noTick)
					{
						m_timeProc = TimerCallback;
					}
					else
					{
						m_timeProc = TimerCallbackNoTick;
					}
				}
				else if (m_mode == TimerMode.PeriodicSemaphore)
				{
					if (!m_noTick)
					{
						m_timeProc = TimerCallbackSemaphore;
					}
					else
					{
						m_timeProc = TimerCallbackSemaphoreNoTick;
					}
				}
				else
				{
					if (!m_noTick)
					{
						m_timeProc = TimerCallbackOneShot;
					}
					else
					{
						m_timeProc = TimerCallbackOneShotNoTick;
					}
				}

				PrecisionTimerSettings.SetMinimumTimerResolution(MMTimerExports.Capabilities.PeriodMinimum);
				m_timerID = timeSetEvent(m_period, m_resolution, m_timeProc, IntPtr.Zero, m_mode == TimerMode.PeriodicSemaphore ? TimerMode.Periodic : m_mode);

				// If the timer was created successfully.
				if (m_timerID != 0)
				{
					m_running = true;

					if (Started is object)
					{
						Started(this, null);
					}

					return true;
				}
				else
				{
					throw new TimerStartException(Constants.unableToStartString);
				}
			}
		}

		/// <summary>
		/// Stop the <see cref="PrecisionTimer"/>
		/// </summary>
		/// <exception cref="ObjectDisposedException">Timer was already Disposed</exception>
		public bool Stop()
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedStopString);
			}

			lock (m_lock)
			{
				if (!m_running)
				{
					return true;
				}

				timeKillEvent(m_timerID);

				m_timerID = 0;
				m_running = false;

				if (Stopped is object)
				{
					Stopped(this, null);
				}
			}

			return true;
		}

		/// <summary>
		/// <para>DESTROY</para>
		/// Release all resources for this <see cref="PrecisionTimer"/>
		/// </summary>
		/// <exception cref="ObjectDisposedException">Timer was already Diposed</exception>
		public void Dispose()
		{
			if (m_disposed)
			{
				throw new ObjectDisposedException(Constants.disposedString);
			}

			Dispose(true);
		}

		protected private virtual void Dispose(bool disposing)
		{
			try
			{
				if (disposing)
				{
					if (m_running)
					{
						Stop();
					}

					Started = null;
					Stopped = null;
					Tick = null;

					m_timerID = 0;
					m_period = 0;
					m_resolution = 0;
					m_timeProc = null;
					m_eventArgs = null;
					m_action = null;
					m_lock = null;

					GC.ReRegisterForFinalize(this);
				}
			}
			finally
			{
				m_disposed = true;
			}
		}

		#region [ Overloads ]

		/// <inheritdoc cref="SetInterval(Action, int, bool, int, TimerMode, bool)"/>
		public void Configure(Action timerTask = null, int interval = 1, bool start = true, int resolution = 0, TimerMode mode = TimerMode.Periodic) => SetInterval(timerTask, interval, start, resolution, mode);

		/// <inheritdoc cref="GetPeriod"/>
		public void GetInterval() => GetPeriod();

		/// <inheritdoc cref="SetPeriod(int)"/>
		public void SetInterval(int value) => SetPeriod(value);

		#endregion [ Overloads ]
	}
}